"""
Mentions for sources that do not support webmentions.
Needs to be input manually.
"""

import logging

from . import MentionsBaseModel
from .mixins.quotable import QuotableMixin


log = logging.getLogger(__name__)


class SimpleMention(QuotableMixin, MentionsBaseModel):
    pass



import logging

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

log = logging.getLogger()


class QuotableMixin(models.Model):
    """A third-party thing that talks about a first-party thing."""

    class Meta:
        abstract = True

    target_url = models.URLField(
        help_text='Our URL that is mentioned')
    source_url = models.URLField(
        help_text='The URL that mentions our content')

    quote = models.CharField(
        max_length=300,
        blank=True,
        null=True,
        help_text='A short excerpt from the quoted piece')

    published = models.DateTimeField(auto_now_add=True)

    hcard = models.ForeignKey(
        'HCard',
        blank=True,
        null=True,
        on_delete=models.CASCADE)

    # The mention target may be an instance of any class that
    # uses the MentionableMixin so we need to use a GenericForeignKey
    content_type = models.ForeignKey(
        ContentType,
        null=True,
        on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField(blank=True, null=True)
    target_object = GenericForeignKey('content_type', 'object_id')



import logging

from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.http import (
    HttpResponse,
    JsonResponse,
    HttpResponseBadRequest,     # HTTP 400
    HttpResponseNotAllowed,     # HTTP 405
)
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from mentions.exceptions import TargetDoesNotExist
from mentions.tasks import process_incoming_webmention
from mentions.util import (
    get_mentions_for_url_path,
    serialize_mentions,
    split_url,
)

log = logging.getLogger(__name__)


def _get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[-1].strip()
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def validate(validator, urls):
    for url in urls:
        try:
            validator(url)
        except ValidationError as e:
            log.warning(f'URL \'{url}\' did not pass validation: {e}')
            return False
    return True


# /webmention/
class WebmentionView(View):
    """Handle incoming webmentions."""

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        log.info('Receiving webmention...')
        if request.method != 'POST':
            return HttpResponseNotAllowed(
                ['POST', ],
                'Only POST requests are accepted')

        http_post = request.POST
        try:
            client_ip = _get_client_ip(request)
            source = http_post['source']
            target = http_post['target']
        except Exception as e:
            log.warning(f'Unable to read webmention params "{http_post}": {e}')
            return HttpResponse(status=400)

        validator = URLValidator(schemes=['http', 'https'])

        if validate(validator, [source, target]):
            log.info(f'Validation passed for source: \'{source}\', target: \'{target}\'')
        else:
            return HttpResponseBadRequest()

        process_incoming_webmention.delay(http_post, client_ip)
        return HttpResponse(status=202)


# /webmention/get
class GetWebmentionsView(View):
    """Return any webmentions associated with a given item."""

    def dispatch(self, request, *args, **kwargs):
        if request.method != 'GET':
            return HttpResponseNotAllowed(['GET', ])

        http_get = request.GET
        for_url = http_get.get('url')

        if not for_url:
            return HttpResponseBadRequest('Missing args')

        scheme, domain, path = split_url(request.build_absolute_uri())
        full_target_url = f'{scheme}://{domain}{for_url}'

        try:
            wm = get_mentions_for_url_path(for_url, full_target_url=full_target_url)
        except TargetDoesNotExist as e:
            log.warning(e)
            return JsonResponse({
                'status': 0,
                'target_url': full_target_url,
                'message': 'Target not found',
            })

        log.info(wm)
        return JsonResponse({
            'status': 1,
            'target_url': full_target_url,
            'mentions': serialize_mentions(wm),
        })



from typing import Tuple
from urllib.parse import urlsplit

import requests
from bs4 import BeautifulSoup
from celery import shared_task
from celery.utils.log import get_task_logger
from django.conf import settings
from django.db import models
from django.http import QueryDict

from mentions.exceptions import (
    BadConfig,
    TargetDoesNotExist,
    TargetWrongDomain,
    SourceNotAccessible,
)
from mentions.models import Webmention, HCard
from mentions.util import (
    get_model_for_url_path,
    split_url,
)

log = get_task_logger(__name__)


class Notes:
    notes = []

    def info(self, note) -> 'Notes':
        log.info(note)
        self.notes.append(note)
        return self

    def warn(self, note) -> 'Notes':
        log.warning(note)
        self.notes.append(note)
        return self

    def __str__(self):
        return '\n'.join(self.notes)


def _update_wm(
        mention,
        target_object = None,
        notes: Notes = None,
        hcard: HCard = None,
        validated: bool = None,
        save: bool = False,
):
    """Update a webmention with the given kwargs"""
    if target_object is not None:
        mention.target_object = target_object
    if notes is not None:
        mention.notes = str(notes)
    if hcard is not None:
        mention.hcard = hcard
    if validated is not None:
        mention.validated = validated

    if save:
        mention.save()
        log.info(f'Webmention saved: {mention}')

    return mention


@shared_task
def process_incoming_webmention(http_post: QueryDict, client_ip: str) -> None:
    log.info(f'Processing webmention \'{http_post}\'')

    # Source and target have already been verified
    # as valid addresses before this method is called
    source = http_post['source']
    target = http_post['target']

    wm = Webmention.create(source, target, sent_by=client_ip)

    # If anything fails, write it to notes and attach to webmention object
    # so it can be checked later
    notes = Notes()

    # Check that the target page is accessible on our server and fetch
    # the corresponding object.
    try:
        obj = _get_target_object(target)
        notes.info('Found webmention target object')
        _update_wm(wm, target_object=obj)
    except (TargetWrongDomain, TargetDoesNotExist):
        notes.warn(f'Unable to find matching page on our server for url {target}')
    except BadConfig:
        notes.warn(f'Unable to find a model associated with url {target}')

    # Verify that the source page exists and really contains a link
    # to the target
    try:
        response_text = _get_incoming_source(source)
    except SourceNotAccessible:
        _update_wm(wm, notes=notes.warn(f'Source not accessible: {source}'), save=True)
        return

    soup = BeautifulSoup(response_text, 'html.parser')
    if not soup.find('a', href=target):
        _update_wm(wm, notes=notes.info('Source does not contain a link to our content'), save=True)
        return

    hcard = HCard.from_soup(soup, save=True)

    _update_wm(wm, validated=True, notes=notes, hcard=hcard, save=True)


def _get_target_object(target_url: str) -> models.Model:
    r"""
    Confirm that the page exists on our server and return object.
    Throws

    Any mentionable Views must include the following kwargs:
        - 'model_name' - dotted python path to the model definition
                         This needs to be explicitely added to kwargs
        - 'slug'       - normally parsed from url pattern, used to
                         resolve the particular model instance

    e.g.
    re_path(
        r'^articles/(?P<slug>[\w\-]+)/?$',
        ArticleView.as_view(),
        kwargs={
            'model_name': 'core.models.Article',
        },
        name='app_view'),

    Raises:
        TargetWrongDomain: If the target_url points to a domain not listed in settings.ALLOWED_HOSTS
        BadConfig: Raised from get_model_for_url
    """
    scheme, domain, path = split_url(target_url)

    if domain not in settings.ALLOWED_HOSTS:
        raise TargetWrongDomain(f'Wrong domain: {domain} (from url={target_url})')

    try:
        return get_model_for_url_path(path)
    except BadConfig as e:
        # log.warning(f'Failed to process incoming webmention! BAD CONFIG: {e}')
        raise e


def _get_incoming_source(source_url: str, client=requests) -> str:
    """
    Fetch the source, confirm content is suitable and return response.
    Verify that the source URL returns an HTML page with a successful
    status code.

    Args:
        source_url: The URL that mentions our content.
        client: A client for making HTTP requests. Should only be explicitly given when testing -
                its API is assumed to be equivalent to `python-requests`.

    Raises:
        SourceNotAccessible: If the `source_url` cannot be resolved, returns an error code, or
                             is an unexpected content type.
    """

    try:
        response = client.get(source_url)
    except Exception as e:
        raise SourceNotAccessible(f'Requests error: {e}')

    if response.status_code >= 300:
        raise SourceNotAccessible(
            f'Source \'{source_url}\' returned error code [{response.status_code}]')

    content_type = response.headers['content-type']
    if 'text/html' not in content_type:
        raise SourceNotAccessible(
            f'Source \'{source_url}\' returned unexpected content type: {content_type}')

    return response.text



import logging

from importlib import import_module
from typing import (
    List,
    Dict,
    Tuple,
)
from urllib.parse import urlsplit

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.urls import (
    Resolver404,
    ResolverMatch,
)

from mentions.exceptions import BadConfig, TargetDoesNotExist
from mentions.models import Webmention

log = logging.getLogger(__name__)


def _find_urlpattern(target_path: str) -> ResolverMatch:
    """
    Find a match in urlpatterns or raise TargetDoesNotExist
    """
    target_path = target_path.lstrip('/')  # Remove any leading slashes
    urlconf = import_module(settings.ROOT_URLCONF)
    urlpatterns = urlconf.urlpatterns
    try:
        from django.contrib.flatpages.views import flatpage
    except ImportError:
        flatpage = None
    for x in urlpatterns:
        # x may be an instance of either:
        # - django.urls.resolvers.URLResolver
        # - django.urls.resolvers.URLPattern
        try:
            match = x.resolve(target_path)
            if match:
                if flatpage and match.func != flatpage:
                    break
                else:
                    break
        except Resolver404:
            # May be raised by URLResolver
            pass
    else:
        # No match found
        raise TargetDoesNotExist(
            f'Cannot find a matching urlpattern entry for path={target_path}')
    log.debug(f'Found matching urlpattern: {match}')
    return match


def split_url(target_url: str) -> Tuple[str, str, str]:
    scheme, full_domain, path, _, _ = urlsplit(target_url)
    domain = full_domain.split(':')[0]  # Remove port number if present
    return scheme, domain, path


def get_model_for_url_path(target_path: str, match: ResolverMatch = None):
    """
    Find a match in urlpatterns and return the corresponding model instance.
    """
    if match is None:
        match = _find_urlpattern(target_path)

    # Dotted path to model class declaration
    model_name = match.kwargs.get('model_name')

    # Required to retrieve the target model instance
    slug = match.kwargs.get('slug')

    if not model_name:
        raise BadConfig(
            'urlpattern must include a kwarg entry called \'model_name\': '
            f'{match}')
    if not slug:
        raise BadConfig(
            f'urlpattern must include a kwarg entry called \'slug\': {match}')

    try:
        model = apps.get_model(model_name)
    except LookupError:
        raise BadConfig(
            f'Cannot find model `{model_name}` - check your urlpattern!')

    try:
        return model.objects.get(slug=slug)
    except ObjectDoesNotExist:
        raise TargetDoesNotExist(
            f'Cannot find instance of model=\'{model}\' with slug=\'{slug}\'')


def get_mentions_for_url_path(target_path: str, full_target_url=None, match: ResolverMatch = None):
    """
    Return a list of webmentions or raise TargetDoesNotExist.
    """
    if match is None:
        match = _find_urlpattern(target_path)

    try:
        obj = get_model_for_url_path(target_path, match)
        return obj.mentions
    except (BadConfig, TargetDoesNotExist):
        pass

    # Add or remove a trailing slash to full_target_url so we can look for both.
    full_target_url_invert_slash = full_target_url[:-1] if full_target_url.endswith('/') else f'{full_target_url}/'

    mentions = Webmention.objects.filter(
        Q(target_url=full_target_url) | Q(target_url=full_target_url_invert_slash) | Q(target_url=target_path),
        approved=True,
        validated=True,
    )

    return mentions


def serialize_mentions(mentions: List[Webmention]) -> List[Dict]:
    return [
        {
            'hcard': mention.hcard.as_json() if mention.hcard else {},
            'quote': mention.quote,
            'source_url': mention.source_url,
            'published': mention.published,
        } for mention in mentions
    ]



import os

TEST_RUNNER = 'django_nose.NoseTestSuiteRunner'
NOSE_ARGS = [
    '--with-spec',
    '--spec-color',
    '--logging-clear-handlers',
    '--traverse-namespace',  # Required since Python 3.8
]

DOMAIN_NAME = 'example.org'
ALLOWED_HOSTS = [
    DOMAIN_NAME,
]

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = 'some-test-key'
INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.flatpages',
    'django.contrib.sites',

    # Apps for django-wm
    'mentions',
    'mentions.tests',
]

MIDDLEWARE = [
    # Middleware for django-wm
    'mentions.middleware.WebmentionHeadMiddleware',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

ROOT_URLCONF = 'mentions.tests.config.test_urls'



"""Models that are only used for testing."""

import logging

from django.db import models
from django.urls import reverse

from mentions.models.mixins.mentionable import MentionableMixin
from mentions.tests.util.constants import view_all_endpoints

log = logging.getLogger(__name__)


class MentionableTestModel(MentionableMixin, models.Model):
    """Stub model for testing."""

    stub_id = models.CharField(max_length=16)
    slug = models.SlugField(unique=True, max_length=20)

    content = models.TextField(blank=True, null=True)

    def get_absolute_url(self):
        abs_url = reverse(view_all_endpoints, args=[self.slug])
        log.info(abs_url)
        return abs_url

    def all_text(self):
        return self.content



import logging

from django.contrib.contenttypes.models import ContentType
from django.db import models

from mentions.models.webmention import Webmention
from mentions.tasks import process_outgoing_webmentions
from mentions.util import serialize_mentions

log = logging.getLogger(__name__)


class MentionableMixin(models.Model):
    class Meta:
        abstract = True

    allow_incoming_webmentions = models.BooleanField(default=True)
    allow_outgoing_webmentions = models.BooleanField(default=False)

    @property
    def mentions(self):
        ctype = ContentType.objects.get_for_model(self.__class__)
        webmentions = Webmention.objects.filter(
            content_type=ctype,
            object_id=self.id,
            approved=True,
            validated=True)
        # manual_mentions = TODO
        return webmentions

    def mentions_json(self):
        return serialize_mentions(self.mentions)

    def all_text(self) -> str:
        """
        Return all the text that should be searched when looking for
        outgoing Webmentions. Any URLs found in this text will be
        checked for webmention support.

        Typically this will just be the main text of your model but
        you may also want to include content from any other text fields
        such as a summary or abstract.

        Example:
            def all_text(self) -> str:
                return f'{self.introduction} {self.main_content}'
        """
        log.warning(
            'This model extends WebMentionableMixin but has not '
            'implemented all_text() so outgoing webmentions will '
            'not work!')
        return ''

    def save(self, *args, **kwargs):
        if self.allow_outgoing_webmentions:
            log.info('Outgoing webmention processing task added to queue...')
            process_outgoing_webmentions.delay(self.get_absolute_url(), self.all_text())

        super().save(*args, **kwargs)



"""
Views that are only used for running tests.
"""

import logging

from django.http import HttpResponse
from django.views.generic.base import View

from mentions.tests.util import snippets
from mentions.tests.util import constants
from mentions.tests.models import MentionableTestModel

log = logging.getLogger(__name__)


class AllEndpointsMentionableTestView(View):
    """
    Expose webmention endpoint in HTTP header, HTML HEAD and HTML BODY.
    """

    def dispatch(self, request, *args, **kwargs):
        obj = MentionableTestModel.objects.get(slug=kwargs.get('slug'))
        html = snippets.html_all_endpoints(obj.content)
        response = HttpResponse(html, status=200)
        response['Link'] = snippets.http_header_link(
            constants.webmention_api_absolute_url)
        return response


class SimpleNoObjectTestView(View):
    """
    A simple view with no associated Mentionable object. Should be able to accept
    incoming webmentions but there is no
    """
    def get(self, request):
        return HttpResponse(snippets.html_all_endpoints('whatever content'), status=200)



import logging

from django.test import TestCase
from django.urls import re_path

from mentions import util
from mentions.exceptions import BadConfig, TargetDoesNotExist
from mentions.tests.util import functions
from mentions.tests.models import MentionableTestModel
from mentions.tests.util import constants
from mentions.tests.views import AllEndpointsMentionableTestView
from mentions.urls import urlpatterns

log = logging.getLogger(__name__)

# Paths as configured in the local urlpatterns
bad_modelname_key = 'with_bad_kwarg_key'
bad_modelname_value = 'with_bad_kwarg_value'
bad_path = 'some_bad_path'

urlpatterns += [
    re_path(
        fr'^{bad_modelname_key}/{constants.slug_regex}',
        AllEndpointsMentionableTestView.as_view(),
        kwargs={
            'model_name_with_mistyped_key': constants.model_name,
        },
        name=f'bad_modelname_key_{constants.view_all_endpoints}'),
    re_path(
        fr'^{bad_modelname_value}/{constants.slug_regex}',
        AllEndpointsMentionableTestView.as_view(),
        kwargs={
            'model_name': 'tests.UnresolvableModel',
        },
        name=f'bad_modelname_value_{constants.view_all_endpoints}'),
]


class GetModelForUrlTests(TestCase):
    """Tests for util.get_model_for_url_path."""

    def setUp(self):
        self.stub_id = 'some-id123'
        self.slug = 'some-slug'

        MentionableTestModel.objects.create(
            stub_id=self.stub_id,
            slug=self.slug).save()

    def test_get_model_for_url__with_unknown_url(self):
        """Ensure that URL with an unrecognised path raises an exception."""
        with self.assertRaises(TargetDoesNotExist):
            util.get_model_for_url_path(functions.url(constants.view_all_endpoints, bad_path))

    def test_get_model_for_url__with_correct_slug(self):
        """Ensure that reverse url lookup finds the correct object."""
        retrieved_object = util.get_model_for_url_path(
            functions.url(constants.correct_config, self.slug))

        self.assertEqual(retrieved_object.stub_id, self.stub_id)

    def test_get_model_for_url__with_unknown_slug(self):
        """Ensure that an unknown slug raises an exception."""
        with self.assertRaises(TargetDoesNotExist):
            util.get_model_for_url_path(functions.url(constants.correct_config, 'unknown-slug'))

    def test_get_model_for_url__with_bad_model_name_config(self):
        """Ensure urlpatterns is set up correctly. urlpatterns must provide a correct dotted path
        value for `model_name` in its kwargs."""

        # Raise BadConfig if kwargs does not have a key named `model_name`.
        with self.assertRaises(BadConfig):
            util.get_model_for_url_path(functions.url(bad_modelname_key, self.slug))

        # Raise BadConfig if `model_name` kwarg has bad/unresolvable path.
        with self.assertRaises(BadConfig):
            util.get_model_for_url_path(functions.url(bad_modelname_value, self.slug))



import json
import logging

from django.db import models

import mf2py

from mentions.models import MentionsBaseModel

log = logging.getLogger(__name__)


class HCard(MentionsBaseModel):
    name = models.CharField(
        blank=True, max_length=50, help_text='Name of the person/organisation')
    avatar = models.URLField(
        blank=True, help_text='Link to their profile image')
    homepage = models.URLField(
        blank=True, null=True, help_text='Link to their homepage')
    json = models.TextField(
        blank=True, help_text='Raw json representation of this hcard')

    def as_json(self):
        return {
            'name': self.name,
            'avatar': self.avatar,
            'homepage': self.homepage,
        }

    @classmethod
    def from_soup(cls, soup, save=False):
        """See https://github.com/microformats/mf2py"""
        parser = mf2py.Parser(doc=soup)
        j = parser.to_dict()
        for item in j['items']:
            if 'h-card' in item.get('type', []):
                log.debug('Reading h-card...')
                attrs = item.get('properties')
                if attrs:
                    _name = attrs.get('name', [''])[0]
                    _avatar = attrs.get('photo', [''])[0]
                    _homepage = attrs.get('url', [''])[0]
                    _json = json.dumps(item, sort_keys=True)

                    if not _name and not _homepage:
                        log.info(f'h-card has neither a name nor a link: {attrs}')
                        continue

                    card, _ = HCard.objects.get_or_create(homepage=_homepage)

                    card.name = _name
                    card.avatar = _avatar
                    card.json = _json

                    if save:
                        card.save()

                    return card
                else:
                    log.info('Could not read "properties"')

    def save(self, *args, **kwargs):
        """Workaround so that empty homepages can be accepted without violating
        'unique' requirement. This allows admin to check missing content manually."""
        if not self.homepage:
            self.homepage = None

        super().save(*args, **kwargs)

    def __str__(self):
        return f'name="{self.name}", avatar="{self.avatar}", homepage="{self.homepage}"'



"""Constants that are used in several separate tests."""
from django.conf import settings
from django.urls import reverse_lazy
import mentions.views.names as view_names

domain = settings.DOMAIN_NAME
namespace = 'webmention'

webmention_api_relative_url = reverse_lazy(view_names.webmention_api_incoming)
webmention_api_get_relative_url = reverse_lazy(
    view_names.webmention_api_get_for_object)

"""Value updated after django initation by runtests.py"""
webmention_api_absolute_url = ''

"""Dotted name of the model used for testing."""
model_name = 'tests.MentionableTestModel'

"""Name of the View used for testing (for reverse lookup)"""
view_all_endpoints = 'all_endpoints_view'
view_no_mentionable_object = 'no_object_view'

"""Path as configured in the local urlpatterns"""
correct_config = 'with_correct_config'

"""Regular expression for the slug used in urlpatterns."""
slug_regex = r'(?P<slug>[\w\-.]+)/?$'



from django.contrib import admin

from .models import (
    HCard,
    SimpleMention,
    Webmention,
    OutgoingWebmentionStatus,
)


def approve_webmention(modeladmin, request, queryset):
    queryset.update(approved=True)


def disapprove_webmention(modeladmin, request, queryset):
    queryset.update(approved=False)


class BaseAdmin(admin.ModelAdmin):
    save_on_top = True


@admin.register(SimpleMention)
class QuotableAdmin(BaseAdmin):
    list_display = [
        'source_url',
        'target_url',
        'hcard',
    ]
    search_fields = [
        'source_url',
        'target_url',
        'hcard',
    ]
    date_hierarchy = 'published'


@admin.register(Webmention)
class WebmentionAdmin(QuotableAdmin):
    readonly_fields = ['target_object', 'published', ]
    actions = [
        approve_webmention,
        disapprove_webmention,
    ]
    list_display = [
        'source_url',
        'target_url',
        'published',
        'validated',
        'approved',
        'content_type',
    ]
    date_hierarchy = 'created_at'
    fieldsets = (
        ('Remote source', {
            'fields': (
                'source_url',
                'sent_by',
                'hcard',
                'quote',
            ),
        }),
        ('Local target', {
            'fields': (
                'target_url',
                'content_type',
                'object_id',
                'target_object',
            ),
        }),
        ('Metadata', {
            'fields': (
                'published',
                'approved',
                'validated',
                'notes',
            ),
        })
    )


@admin.register(OutgoingWebmentionStatus)
class OutgoingWebmentionStatusAdmin(BaseAdmin):
    readonly_fields = [
        'created_at',
        'source_url',
        'target_url',
        'target_webmention_endpoint',
        'status_message',
        'response_code',
        'successful',
    ]
    list_display = [
        'source_url',
        'target_url',
        'successful',
        'created_at',
    ]
    date_hierarchy = 'created_at'


@admin.register(HCard)
class HCardAdmin(BaseAdmin):
    list_display = ['name', 'homepage']
    search_fields = ['name', 'homepage']



"""
Utility functions used in multiple test files.
"""

import uuid

from django.utils.text import slugify
from django.urls import reverse

from . import constants


def get_id_and_slug():
    """Create a random id and slug for a MentionableTestModel."""
    _id = uuid.uuid4().hex[:6]
    slug = slugify(_id)
    return _id, slug


def url(path, slug=''):
    return f'/{constants.namespace}/{path}/{slug}'


def get_mentioning_content(_url):
    """Return html content that links to the given url."""
    return f'''This is some content that mentions the target <a href="{_url}">url</a>'''


def build_object_url(slug):
    return f'https://{constants.domain}{reverse(constants.view_all_endpoints, args=[slug])}'


def build_object_relative_url(slug):
    return reverse(constants.view_all_endpoints, args=[slug])


"""
Tests for webmentions that originate on our server, usually pointing somewhere else.
"""
import logging
from unittest import mock

import requests
from django.test import TestCase
from django.urls import reverse

from mentions.models import OutgoingWebmentionStatus
from mentions.tasks.outgoing_webmentions import _send_webmention
from mentions.tests.util import snippets
from mentions.tasks import (
    outgoing_webmentions,
    process_outgoing_webmentions,
)
from mentions.tests.util import functions
from mentions.tests.models import MentionableTestModel
from mentions.tests.util import constants

log = logging.getLogger(__name__)


OUTGOING_WEBMENTION_HTML = """<html>
<head><link rel="webmention" href="/webmention/" /></head>
<body>blah blah 
<a href="https://django-wm.dev/">This is a mentionable link</a> 
blah blah</body></html>
"""

OUTGOING_WEBMENTION_HTML_MULTIPLE_LINKS = """<html>
<head><link rel="webmention" href="/webmention/" /></head>
<body>blah blah 
<a href="https://django-wm.dev/">This is a mentionable link</a> 
<a href="https://django-wm.dev/some-article/">This is another mentionable link</a> 
blah blah</body></html>
"""

OUTGOING_WEBMENTION_HTML_NO_LINKS = """<html>
<head><link rel="webmention" href="/webmention/" /></head>
<body>blah blah no links here blah blah</body></html>
"""


class MockResponse:
    """Mock of requests.Response."""

    def __init__(
        self,
        url: str,
        headers: dict = None,
        text: str = None,
        status_code: int = None,
        *args,
        **kwargs,
    ):
        if headers is None:
            headers = {}

        self.url = url
        self.text = text
        self.status_code = status_code
        self.headers = headers

        if args:
            print(f"Unhandled args: {args}")
            self.args = args

        if kwargs:
            print(f"Unhandled kwargs: {kwargs}")
            self.kwargs = kwargs


def _get_mock_get_response_ok(url, *args, **kwargs):
    return {
        "https://django-wm.dev/": MockResponse(
            "https://django-wm.dev/",
            text=OUTGOING_WEBMENTION_HTML,
            headers=kwargs.get("headers", {}),
            status_code=200,
        ),
        "https://django-wm.dev/some-article/": MockResponse(
            "https://django-wm.dev/some-article/",
            text=OUTGOING_WEBMENTION_HTML_MULTIPLE_LINKS,
            headers=kwargs.get("headers", {}),
            status_code=200,
        ),
    }.get(url)


def _get_mock_post_response_ok(url, *args, **kwargs):
    return {
        "https://django-wm.dev/webmention/": MockResponse(
            "https://django-wm.dev/webmention/",
            headers=kwargs.get("headers", {}),
            status_code=200,
        ),
    }.get(url)


def _get_mock_post_response_endpoint_error(url, *args, **kwargs):
    return {
        "https://django-wm.dev/webmention/": MockResponse(
            "https://django-wm.dev/webmention/",
            text=OUTGOING_WEBMENTION_HTML,
            headers=kwargs.get("headers", {}),
            status_code=400,
        ),
    }.get(url)


class OutgoingWebmentionsTests(TestCase):
    """"""

    def setUp(self):
        self.target_stub_id, self.target_slug = functions.get_id_and_slug()

        target = MentionableTestModel.objects.create(
            stub_id=self.target_stub_id,
            slug=self.target_slug,
            allow_incoming_webmentions=True,
        )
        target.save()

    def _get_target_url(self, viewname):
        return reverse(viewname, args=[self.target_slug])

    def _get_absolute_target_url(self, viewname):
        return f"https://{constants.domain}{self._get_target_url(viewname)}"

    def test_find_links_in_text(self):
        """Ensure that outgoing links are found correctly."""

        outgoing_content = functions.get_mentioning_content(
            reverse(constants.view_all_endpoints, args=[self.target_slug])
        )

        outgoing_links = outgoing_webmentions._find_links_in_text(outgoing_content)
        self.assertEqual(
            outgoing_links, [functions.url(constants.correct_config, self.target_slug)]
        )

    def test_get_absolute_endpoint_from_response(self):
        """Ensure that any exposed endpoints are found and returned as an absolute url."""
        mock_response = MockResponse(
            url=self._get_absolute_target_url(constants.view_all_endpoints),
            headers={"Link": snippets.http_link_endpoint()},
        )
        absolute_endpoint_from_http_headers = outgoing_webmentions._get_absolute_endpoint_from_response(
            mock_response
        )
        self.assertEqual(
            constants.webmention_api_absolute_url, absolute_endpoint_from_http_headers
        )

        mock_response.headers = {}
        mock_response.text = snippets.html_head_endpoint()
        absolute_endpoint_from_html_head = outgoing_webmentions._get_absolute_endpoint_from_response(
            mock_response
        )
        self.assertEqual(
            constants.webmention_api_absolute_url, absolute_endpoint_from_html_head
        )

        mock_response.headers = {}
        mock_response.text = snippets.html_body_endpoint()
        absolute_endpoint_from_html_body = outgoing_webmentions._get_absolute_endpoint_from_response(
            mock_response
        )
        self.assertEqual(
            constants.webmention_api_absolute_url, absolute_endpoint_from_html_body
        )

    def test_get_endpoint_in_http_headers(self):
        """Ensure that endpoints exposed in http header are found correctly."""

        mock_response = MockResponse(
            url=self._get_absolute_target_url(constants.view_all_endpoints),
            headers={"Link": snippets.http_link_endpoint()},
        )
        endpoint_from_http_headers = outgoing_webmentions._get_endpoint_in_http_headers(
            mock_response
        )
        self.assertEqual(
            constants.webmention_api_relative_url, endpoint_from_http_headers
        )

    def test_get_endpoint_in_html_head(self):
        """Ensure that endpoints exposed in html <head> are found correctly."""

        mock_response = MockResponse(
            url=self._get_absolute_target_url(constants.view_all_endpoints),
            text=snippets.html_head_endpoint(),
        )
        endpoint_from_html_head = outgoing_webmentions._get_endpoint_in_html(
            mock_response
        )
        self.assertEqual(constants.webmention_api_relative_url, endpoint_from_html_head)

    def test_get_endpoint_in_html_body(self):
        """Ensure that endpoints exposed in html <body> are found correctly."""

        mock_response = MockResponse(
            url=self._get_absolute_target_url(constants.view_all_endpoints),
            text=snippets.html_body_endpoint(),
        )
        endpoint_from_html_body = outgoing_webmentions._get_endpoint_in_html(
            mock_response
        )
        self.assertEqual(constants.webmention_api_relative_url, endpoint_from_html_body)

    def test_relative_to_absolute_url(self):
        """Ensure that relative urls are correctly converted to absolute"""

        domain = constants.domain
        page_url = f"https://{domain}/some-url-path"
        response = MockResponse(url=page_url)

        absolute_url_from_root = outgoing_webmentions._relative_to_absolute_url(
            response, "/webmention/"
        )
        self.assertEqual(f"https://{domain}/webmention/", absolute_url_from_root)

        absolute_url_from_relative = outgoing_webmentions._relative_to_absolute_url(
            response, "webmention/"
        )
        self.assertEqual(f"https://{domain}/webmention/", absolute_url_from_relative)

        already_absolute_url = outgoing_webmentions._relative_to_absolute_url(
            response, f"https://{domain}/already_absolute_path"
        )
        self.assertEqual(
            f"https://{domain}/already_absolute_path", already_absolute_url
        )

    @mock.patch.object(
        requests, "post", mock.Mock(side_effect=_get_mock_post_response_ok,)
    )
    def test_send_webmention(self):
        page_url = f"https://{constants.domain}/some-url-path/"
        success, status_code = _send_webmention(
            source_url=page_url,
            endpoint="https://django-wm.dev/webmention/",
            target="https://django-wm.dev/",
        )
        self.assertTrue(success)
        self.assertEqual(200, status_code)

    @mock.patch.object(
        requests, "post", mock.Mock(side_effect=_get_mock_post_response_endpoint_error,)
    )
    def test_send_webmention__with_endpoint_error(self):
        page_url = f"https://{constants.domain}/some-url-path/"
        success, status_code = _send_webmention(
            source_url=page_url,
            endpoint="https://django-wm.dev/webmention/",
            target="https://django-wm.dev/",
        )
        self.assertFalse(success)
        self.assertEqual(400, status_code)

    @mock.patch.object(
        requests, "get", mock.Mock(side_effect=_get_mock_get_response_ok,)
    )
    @mock.patch.object(
        requests, "post", mock.Mock(side_effect=_get_mock_post_response_ok,)
    )
    def test_process_outgoing_webmentions(self):
        """Test the entire process_outgoing_webmentions task."""
        self.assertEqual(0, OutgoingWebmentionStatus.objects.count())
        page_url = f"https://{constants.domain}/some-url-path/"
        page_text = OUTGOING_WEBMENTION_HTML

        wm = process_outgoing_webmentions(page_url, page_text)

        self.assertEqual(1, wm)
        self.assertEqual(1, OutgoingWebmentionStatus.objects.count())

        page_text = OUTGOING_WEBMENTION_HTML_MULTIPLE_LINKS
        wm = process_outgoing_webmentions(page_url, page_text)

        self.assertEqual(2, wm)
        self.assertEqual(3, OutgoingWebmentionStatus.objects.count())

    @mock.patch.object(
        requests,
        "get",
        mock.Mock(
            return_value=None,  # No network requests should be made if links not found in text
        ),
    )
    @mock.patch.object(
        requests,
        "post",
        mock.Mock(
            return_value=None,  # No network requests should be made if links not found in text
        ),
    )
    def test_process_outgoing_webmentions__with_no_links_found(self):
        """Test the entire process_outgoing_webmentions task with no links in provided text."""
        self.assertEqual(0, OutgoingWebmentionStatus.objects.count())
        page_url = f"https://{constants.domain}/some-url-path/"
        page_text = OUTGOING_WEBMENTION_HTML_NO_LINKS

        wm = process_outgoing_webmentions(page_url, page_text)

        self.assertEqual(0, wm)
        self.assertEqual(0, OutgoingWebmentionStatus.objects.count())

    @mock.patch.object(
        requests, "get", mock.Mock(side_effect=_get_mock_get_response_ok,)
    )
    @mock.patch.object(
        requests, "post", mock.Mock(side_effect=_get_mock_post_response_endpoint_error,)
    )
    def test_process_outgoing_webmentions__with_endpoint_error(self):
        """Test the entire process_outgoing_webmentions task."""
        self.assertEqual(0, OutgoingWebmentionStatus.objects.count())
        page_url = f"https://{constants.domain}/some-url-path/"
        page_text = OUTGOING_WEBMENTION_HTML

        wm = process_outgoing_webmentions(page_url, page_text)

        self.assertEqual(0, wm)
        self.assertEqual(1, OutgoingWebmentionStatus.objects.count())



class TargetWrongDomain(Exception):
    """Target URL does not point to any domain in settings.ALLOWED_HOSTS."""

    pass


class TargetDoesNotExist(Exception):
    """Target URL does not point to an object on our server."""

    pass


class SourceNotAccessible(Exception):
    """Source URL does not exist, or returns an error code."""

    pass


class SourceDoesNotLink(Exception):
    """Source URL exists but does not contain link to our content."""

    pass


class BadConfig(Exception):
    """
    URL resolution completed but did not include required data.

    The returned ResolverMatch.kwargs object must have entries for:
        - 'model_name'
        - 'slug'
    """

    pass



"""
Make sure we can correctly retrieve Webmentions for a given url/object.
"""

import json
import logging

from django.test import TestCase
from django.urls import reverse

from mentions.models import Webmention
from mentions.tests.models import MentionableTestModel
from mentions.tests.util import functions
from mentions.tests.util import constants

log = logging.getLogger(__name__)


def _create_quick_webmention(source, target, valid=True, approved=True):
    wm = Webmention.create(
        functions.build_object_url(source.slug),
        functions.build_object_url(target.slug),
        sent_by='tests@localhost')
    wm.validated = valid
    wm.approved = approved
    wm.target_object = target
    wm.save()


class WebmentionGetTests(TestCase):
    """"""
    def setUp(self):
        self.target_stub_id, self.target_slug = functions.get_id_and_slug()

        target = MentionableTestModel.objects.create(
            stub_id=self.target_stub_id,
            slug=self.target_slug,
            allow_incoming_webmentions=True)
        target.save()
        self.target_url = functions.build_object_relative_url(self.target_slug)

        self.source_stub_id, self.source_slug = functions.get_id_and_slug()
        source = MentionableTestModel.objects.create(
            stub_id=self.source_stub_id,
            slug=self.source_slug,
            content=functions.get_mentioning_content(self.target_url))
        source.save()
        self.source_url = functions.build_object_url(self.source_slug)

        _create_quick_webmention(source, target)

    def test_get_webmentions_view__with_valid_object(self):
        """Ensure that we can correctly retrieve webmentions for a given url."""
        response = self.client.get(
            constants.webmention_api_get_relative_url,
            data={'url': self.target_url})

        json_response = json.loads(response.content)
        log.info(json_response)
        self.assertEqual(json_response['status'], 1)

        _mentions = json_response['mentions']
        self.assertEqual(len(_mentions), 1)

        first_mention = _mentions[0]
        self.assertEqual(first_mention['source_url'], self.source_url)


class WebmentionGetBadRequestTests(TestCase):
    """"""
    def test_get_webmentions_view__require_http_get(self):
        """Ensure that webmentions get view does not accept HTTP POST."""
        response = self.client.post(constants.webmention_api_get_relative_url)
        self.assertEqual(405, response.status_code)

    def test_get_webmentions_view__require_param_for_url(self):
        response = self.client.get(constants.webmention_api_get_relative_url)
        self.assertEqual(400, response.status_code)

    def test_get_webmentions_view__target_does_not_exist(self):
        response = self.client.get(
            constants.webmention_api_get_relative_url,
            data={'url': '/does-not-exist'})

        json_response = json.loads(response.content)
        self.assertEqual(json_response['status'], 0)



class WebmentionNoModelTests(TestCase):
    """"""
    def setUp(self) -> None:
        self.target_url = reverse(constants.view_no_mentionable_object)
        Webmention.objects.create(
            source_url='https://django-wm.dev/',
            target_url=self.target_url,
            approved=True,
            validated=True,
        ).save()

    def test_get_webmentions_view__no_mentionable_model(self):
        response = self.client.get(
            constants.webmention_api_get_relative_url,
            data={'url': self.target_url})

        json_response = json.loads(response.content)
        print(json_response)

        _mentions = json_response['mentions']
        self.assertEqual(len(_mentions), 1)

        first_mention = _mentions[0]
        self.assertEqual(first_mention['source_url'], 'https://django-wm.dev/')



from django.urls import path, re_path

import mentions.views.names as view_names
from mentions.views.webmention import GetWebmentionsView, WebmentionView

"""
/webmention/
Configure in root project urls.py
e.g.
    from django.conf import settings
    urlpatterns = [
        ...,
        path(r'webmention/', include('mentions.urls')),
        ...,
    ]
"""
urlpatterns = [
    path(
        '',
        WebmentionView.as_view(),
        name=view_names.webmention_api_incoming),
    re_path(
        r'^get/?$',
        GetWebmentionsView.as_view(),
        name=view_names.webmention_api_get_for_object),
]



from django.urls import path, include, re_path

from mentions.tests.util import constants
from mentions.tests.views import (
    AllEndpointsMentionableTestView,
    SimpleNoObjectTestView,
)

urlpatterns = [
    re_path(
        fr'^{constants.namespace}/{constants.correct_config}/{constants.slug_regex}',
        AllEndpointsMentionableTestView.as_view(),
        kwargs={
            'model_name': constants.model_name,
        },
        name=constants.view_all_endpoints),
    path(f'{constants.namespace}/', include('mentions.urls')),
    path('some-page/', SimpleNoObjectTestView.as_view(), name=constants.view_no_mentionable_object)
]



"""
Ensure HCard objects are correctly built from HTML elements with the 'h-card' class.
"""

import logging

from bs4 import BeautifulSoup
from django.test import TestCase

from mentions.models import HCard

log = logging.getLogger(__name__)


class HCardTests(TestCase):
    """"""
    def test_hcard__with_name_homepage_noavatar(self):
        sample = '''<a class="h-card" href="http://joebloggs.com">Joe Bloggs</a>'''
        beautifulsoup = BeautifulSoup(sample, features='html5lib')
        hcard = HCard.from_soup(beautifulsoup)

        self.assertEqual(hcard.name, 'Joe Bloggs')
        self.assertEqual(hcard.homepage, 'http://joebloggs.com')
        self.assertFalse(hcard.avatar)  # Empty

    def test_hcard__with_name_avatar_image(self):
        sample = '''<p class="h-card"><img class="u-photo" src="https://janebloggs.org/photo.jpg" alt="" /><a class="p-name u-url" href="https://janebloggs.org">Jane Bloggs</a><a class="u-email" href="mailto:janebloggs@janebloggs.com">janebloggs@janebloggs.com</a>,<span class="p-street-address">17 Austerstræti</span><span class="p-locality">Reykjavík</span><span class="p-country-name">Iceland</span></p>'''
        beautifulsoup = BeautifulSoup(sample, features='html5lib')
        hcard = HCard.from_soup(beautifulsoup)

        self.assertEqual(hcard.name, 'Jane Bloggs')
        self.assertEqual(hcard.homepage, 'https://janebloggs.org')
        self.assertEqual(hcard.avatar, 'https://janebloggs.org/photo.jpg')

    def test_hcard__with_beatonma_hcard(self):
        """Test my own h-card from beatonma.org."""
        sample = '''<div id="some_container"><div class="h-card flex-row-center footer"><a rel="me" href="https://beatonma.org" class="u-url foot-link" alt="beatonma.org" aria-label="beatonma.org"><div class="tooltip"><svg id="footer_icon_home" class="foot-icon" viewBox="0 0 48 48"><path d="M 19.6,42.8 L 19.6,29.5 L 28.5,29.5 L 28.5,42.8 L 39.6,42.8 L 39.6,25 L 46.3,25 L 24,5.05 L 1.75,25 L 8.45,25 L 8.45,42.8 Z"/></svg><div class="tooltip-popup">beatonma.org</div></div></a><a rel="me" href="https://play.google.com/store/apps/developer?id=Michael+Beaton" class="foot-link" alt="Play Store" aria-label="Play Store"><div class="tooltip"><svg id="footer_icon_playstore" class="foot-icon" viewBox="0 0 48 48"><path d="M 32.5,30.6 L 6.66,4.71 M 32.5,17.4 L 6.72,43.3 M 32.5,30.6 C 45,23.2 44.5,24.4 32.5,17.4 L 9.92,4.76 C 9.02,4.19 7.39,3.53 6.66,4.71 C 6.48,5.01 6.35,5.42 6.31,5.99 L 6.34,42 C 6.34,42.6 6.48,43 6.72,43.3 C 7.41,44.2 8.92,44 10.1,43.4 Z"/></svg><div class="tooltip-popup">Play Store</div></div></a><a rel="me" href="https://github.com/beatonma" class="foot-link" alt="Github" aria-label="Github"><div class="tooltip"><svg id="footer_icon_github" class="foot-icon" viewBox="0 0 48 48"><path d="M 24 3.95 C 12.7 3.95 3.45 13.1 3.45 24.6 C 3.45 33.6 9.35 41.2 17.5 43.9 C 18.6 44.1 18.9 43.5 18.9 43 C 18.9 42.5 18.9 41.2 18.9 39.5 C 13.2 40.7 12 36.7 12 36.7 C 11 34.3 9.65 33.7 9.65 33.7 C 7.85 32.5 9.89 32.5 9.89 32.5 C 11.9 32.7 13 34.6 13 34.6 C 14.8 37.7 17.8 36.8 19 36.3 C 19.1 34.9 19.7 34 20.3 33.6 C 15.7 33.1 10.9 31.4 10.9 23.4 C 10.9 21.3 11.7 19.4 13 18 C 12.8 17.5 12.1 15.4 13.2 12.5 C 13.2 12.5 15 11.9 18.9 14.6 C 20.5 14.1 22.3 13.8 24 13.8 C 25.8 13.8 27.5 14.1 29.2 14.6 C 33 11.9 34.8 12.5 34.8 12.5 C 36 15.4 35.2 17.5 35 18 C 36.3 19.4 37.1 21.3 37.1 23.4 C 37.1 31.4 32.3 33.1 27.7 33.6 C 28.5 34.2 29.1 35.4 29.1 37.3 C 29.1 40.1 29.1 42.3 29.1 43 C 29.1 43.5 29.5 44.1 30.6 43.9 C 38.6 41.2 44.5 33.6 44.5 24.6 C 44.5 13.1 35.3 3.95 24 3.95"/></svg><div class="tooltip-popup">Github</div></div></a><a rel="me" href="https://twitter.com/_beatonma" class="foot-link" alt="Twitter" aria-label="Twitter"><div class="tooltip"><svg id="footer_icon_twitter" class="foot-icon" viewBox="0 0 56 48"><path d="M 18.7 44 C 37.2 44 47.7 29 47.7 15 C 47.7 15 47.7 14 47.7 14 C 49.7 13 51.7 11 52.7 8.95 C 50.7 9.95 48.7 9.95 46.7 9.95 C 48.7 8.95 50.7 6.95 51.7 4.95 C 49.7 5.95 46.7 6.95 44.7 6.95 C 42.7 4.95 40.2 3.95 37.4 3.95 C 31.7 3.95 27.2 8.95 27.2 14 C 27.2 15 27.4 16 27.6 16 C 19.1 16 11.6 12 6.7 5.95 C 5.8 6.95 5.3 8.95 5.3 11 C 5.3 14 7.1 18 9.8 19 C 8.2 19 6.6 19 5.3 18 C 5.3 18 5.3 18 5.3 18 C 5.3 23 8.8 27 13.4 28 C 12.4 28 11.6 28 10.7 28 C 10.1 28 9.4 28 8.8 28 C 10.1 32 13.8 35 18.3 35 C 14.8 38 10.4 40 5.7 40 C 4.9 40 4.1 40 3.3 39 C 7.7 42 13 44 18.7 44"/></svg><div class="tooltip-popup">Twitter</div></div></a><a rel="me" href="/feed" class="foot-link noanim" alt="RSS feed" aria-label="RSS feed"><div class="tooltip"><svg id="footer_icon_feed" class="foot-icon" viewBox="0 0 48 48"><circle cx="9.6" cy="38.3" r="5.59"/><path d="M 4 3.95 L 4 11.4 C 21.9 11.4 36.6 26.1 36.6 44 L 44 44 C 44 21.9 26.1 3.95 4 3.95 Z M 4 18.6 L 4 25.8 C 14 25.8 22.2 34 22.2 44 L 29.4 44 C 29.4 29.9 18.1 18.6 4 18.6 Z"/></svg><div class="tooltip-popup">RSS feed</div></div></a><div class="nodisplay"><span class="p-name">Michael Beaton</span><span class="p-region">Scotland</span><span class="p-country-name">UK</span><div class="dt-bday">1987-05-16</div><img class="u-logo u-photo" src="https://beatonma.org/static/images/mb.png" /><a rel="me" href="https://www.last.fm/user/schadenfreude87" aria-label="Last.fm profile page">My listening habits since 2005</a><a rel="me" href="https://userstyles.org/users/365999" aria-label="Userstyles.org profile page">Custom CSS style overrides</a><a rel="me" href="http://eu.battle.net/sc2/en/profile/2784180/1/fallofmath/" aria-label="StarCraft II battle.net profile page">StarCraft II</a><a rel="me" href="https://www.duolingo.com/fallofmath" aria-label="Duolingo profile page">Duolingo</a><a rel="me" href="https://gravatar.com/beatonma" aria-label="Gravatar profile page">Gravatar</a><a rel="webmention" href="https://beatonma.org/webmention/" aria-label="Webmention endpoint for beatonma.org" /></div></div></div>'''
        beautifulsoup = BeautifulSoup(sample, features='html5lib')
        hcard = HCard.from_soup(beautifulsoup)

        self.assertEqual(hcard.name, 'Michael Beaton')
        self.assertEqual(hcard.homepage, 'https://beatonma.org')
        self.assertEqual(hcard.avatar, 'https://beatonma.org/static/images/mb.png')

    def test_hcard__with_no_hcard_element(self):
        """Ensure that HTML without h-card an element does not produce an HCard object."""
        sample = '''<div id="some_container"><div class="not-an-h-card"><div class="fake-name">Steve</div></div></div>'''
        beautifulsoup = BeautifulSoup(sample, features='html5lib')
        hcard = HCard.from_soup(beautifulsoup)

        self.assertIsNone(hcard)



from django.conf import settings
from django.urls import reverse
from mentions.views.names import webmention_api_incoming


class WebmentionHeadMiddleware:
    """Automatically add webmention endpoint to HTTP headers of any HttpResponse
    objects created by this server."""
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        port = request.META.get('SERVER_PORT', 80)
        path = reverse(webmention_api_incoming)
        response['Link'] = (
            f'<{request.scheme}://{settings.DOMAIN_NAME}:{port}{path}>'
            ';rel="webmention"')
        return response



"""

"""

import logging

from django.db import models

log = logging.getLogger(__name__)


class MentionsBaseModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, null=True)

    class Meta:
        abstract = True



import logging

from django.conf import settings
from django.db import models

from .base import MentionsBaseModel
from .mixins.quotable import QuotableMixin

log = logging.getLogger(__name__)


def _approve_default():
    if hasattr(settings, 'WEBMENTIONS_AUTO_APPROVE'):
        return settings.WEBMENTIONS_AUTO_APPROVE
    return False


class Webmention(QuotableMixin, MentionsBaseModel):
    """An incoming webmention that is received by your server."""
    sent_by = models.URLField(
        blank=True,
        help_text='Source address of the HTTP request that sent this '
                  'webmention')

    approved = models.BooleanField(
        default=_approve_default,
        help_text='Allow this webmention to appear publicly')
    validated = models.BooleanField(
        default=False,
        help_text='True if both source and target have been validated, '
                  'confirmed to exist, and source really does link to target')

    notes = models.CharField(max_length=1024, blank=True)

    @classmethod
    def create(cls, source_url, target_url, sent_by=None):
        return cls(source_url=source_url, target_url=target_url, sent_by=sent_by)

    def approve(self):
        self.approved = True

    class Meta:
        ordering = ['-created_at']
        permissions = [
            ('approve_webmention', 'Can approve received Webmentions for publishing.')
        ]

    def __str__(self):
        return (f'{self.source_url} -> {self.target_url} '
                f'[validated={self.validated}, approved={self.approved},'
                f'content_type={self.content_type}, id={self.object_id}]')


class OutgoingWebmentionStatus(MentionsBaseModel):
    """Status tracker for webmentions that you (attempt to) send from your server.

    Currently used primarily for logging of outgoing mentions.
    """
    source_url = models.URLField(
        help_text='The URL on your server where this mention originates',
    )
    target_url = models.URLField(
        help_text='The URL that you mentioned.',
    )
    target_webmention_endpoint = models.URLField(
        null=True,
        blank=True,
        help_text='The endpoint URL to which we sent the webmention',
    )
    status_message = models.CharField(
        max_length=1024,
        help_text='Success, or an explanation of what went wrong.',
    )
    response_code = models.PositiveIntegerField(default=0)

    successful = models.BooleanField(default=False)

    def __str__(self):
        return (f'{self.source_url} -> {self.target_url} '
                f'(endpoint={self.target_webmention_endpoint}): '
                f'[{self.successful}] {self.status_message} '
                f'[{self.response_code}]')

    class Meta:
        ordering = ['-created_at']
        verbose_name_plural = 'Outgoing Webmention Statuses'



import os
from setuptools import find_packages, setup
import mentions

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

setup(
    name='django-wm',
    version=mentions.__version__,
    packages=find_packages(),
    include_package_data=True,
    license='GPLv3',
    description='Webmention support for any Django model.',
    long_description=README,
    long_description_content_type='text/x-rst',
    url='https://beatonma.org/django-wm',
    project_urls={
        'Source': 'https://github.com/beatonma/django-wm',
        'Implementation tester': 'https://django-wm.dev',
        'Example implementation source': 'https://github.com/beatonma/django-wm-example'
    },
    author='Michael Beaton',
    author_email='michael@beatonma.org',
    install_requires=[
        'beautifulsoup4',
        'celery',
        'django',
        'mf2py',
        'requests',
    ],
    python_requires='>=3.6',
    tests_require=[
        'django-nose',
        'pinocchio',
    ],
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



"""
Chunks of markup.
"""

import logging

from mentions.tests.util import constants

log = logging.getLogger(__name__)


def http_header_link(url):
    return f'<{url}>; rel="webmention"'


def html_head_link(url):
    return f'<link rel="webmention" href="{url}"/>'


def html_body_link(url):
    return f'<a rel="webmention" href="{url}">Endpoint!</a>'


def http_link_endpoint():
    return http_header_link(constants.webmention_api_relative_url)

def html_head_endpoint():
    return f'''
<html>
<head>
    {html_head_link(constants.webmention_api_relative_url)}
</head>
<body></body>
</html>
'''

def html_body_endpoint():
    return f'''
<html>
<head></head>
<body>
    {html_body_link(constants.webmention_api_relative_url)}
</body>
</html>
'''

def html_all_endpoints(content):
    return f'''
    <html>
    <head>
    {html_head_link(constants.webmention_api_relative_url)}
    </head>
    <body>
    <div>
    {content}
    </div>
    <div>
    {html_body_link(constants.webmention_api_relative_url)}
    </div>
    </body>
    </html>
'''



import logging
import re
from typing import (
    Optional,
    Tuple,
    List,
)
from urllib.parse import urlsplit

from django.conf import settings

import requests

from bs4 import BeautifulSoup
from celery import shared_task
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

from mentions.models import OutgoingWebmentionStatus

log = logging.getLogger(__name__)


STATUS_MESSAGE_TARGET_UNREACHABLE = 'The target URL could not be retrieved.'
STATUS_MESSAGE_TARGET_ERROR_CODE = 'The target URL returned an HTTP error code.'
STATUS_MESSAGE_TARGET_ENDPOINT_ERROR = 'The target endpoint URL returned an HTTP error code'
STATUS_MESSAGE_OK = 'The target server accepted the webmention.'


def _save_status(
        outgoing_status: OutgoingWebmentionStatus,
        message: str,
        success: bool,
        target_endpoint: Optional[str] = None,
        response_code: Optional[int] = None,
):
    outgoing_status.status_message = message
    outgoing_status.successful = success
    if target_endpoint:
        outgoing_status.target_webmention_endpoint = target_endpoint
    if response_code:
        outgoing_status.response_code = response_code
    outgoing_status.save()


@shared_task
def process_outgoing_webmentions(source_url: str, text: str) -> int:
    """
    Find links and send webmentions if remote server accepts them.

    Returns True if submissions is successful, False if submission fails, or
    None if no endpoints were found.


    Spec:
    https://www.w3.org/TR/webmention/#sender-discovers-receiver-webmention-endpoint

    For each link found in text:
        - Check that the link is alive
        - Search for webmention endpoint in:
            - HTTP response headers
            - HTML <head> element
            - HTML <body> element
        - If an endpoint is found, send a webmention notification

    source_url should be the value returned by model.get_absolute_url() -
    it will be appended to settings.DOMAIN_NAME
    """
    log.info(f'Checking for outgoing webmention links...')
    mentions_sent = 0
    links_in_text = _find_links_in_text(text)

    if not links_in_text:
        log.info('No webmentionable links found in text.')
        log.debug(f'No webmentionable links in text [{text}]')
        return 0

    for link_url in links_in_text:
        outgoing_status = OutgoingWebmentionStatus.objects.create(
            source_url=source_url,
            target_url=link_url,
        )
        # Confirm that the target url is alive
        log.info(f'Checking url={link_url}')
        try:
            response = requests.get(link_url)
        except Exception as e:
            log.warning(f'Unable to fetch url={link_url}: {e}')
            _save_status(outgoing_status, STATUS_MESSAGE_TARGET_UNREACHABLE, success=False)
            continue

        if response.status_code >= 300:
            log.warning(
                f'Link "{link_url}" returned status={response.status_code}')
            _save_status(
                outgoing_status,
                STATUS_MESSAGE_TARGET_ERROR_CODE,
                success=False,
                response_code=response.status_code,
            )
            continue

        endpoint = _get_absolute_endpoint_from_response(response)
        if endpoint:
            log.info(f'Found webmention endpoint: {endpoint}')
            success, status_code = _send_webmention(source_url, endpoint, link_url)
            if success:
                mentions_sent += 1
                _save_status(
                    outgoing_status,
                    STATUS_MESSAGE_OK,
                    target_endpoint=endpoint,
                    success=success,
                    response_code=status_code,
                )
            else:
                _save_status(
                    outgoing_status,
                    STATUS_MESSAGE_TARGET_ENDPOINT_ERROR,
                    target_endpoint=endpoint,
                    success=success,
                    response_code=status_code,
                )
        else:
            log.info(f'No webmention endpoint found for url {link_url}')

    if mentions_sent:
        log.info(f'Successfully sent {mentions_sent} webmentions')
    else:
        log.info(f'No webmentions were successful!')

    return mentions_sent


def _find_links_in_text(text: str) -> List[str]:
    soup = BeautifulSoup(text, 'html.parser')
    return [a['href'] for a in soup.find_all('a', href=True)]


def _get_absolute_endpoint_from_response(response: requests.Response) -> Optional[str]:
    endpoint = (_get_endpoint_in_http_headers(response) or
                _get_endpoint_in_html(response))
    abs_url = _relative_to_absolute_url(response, endpoint)
    log.debug(f'Absolute url: {endpoint} -> {abs_url}')
    return abs_url


def _get_endpoint_in_http_headers(response: requests.Response) -> Optional[str]:
    """Search for webmention endpoint in HTTP headers."""
    try:
        header_link = response.headers.get('Link')
        if 'webmention' in header_link:
            log.debug('webmention endpoint found in http header')
            endpoint = re.match(
                r'<(?P<url>.*)>[; ]*.rel=[\'"]?webmention[\'"]?',
                header_link).group(1)
            return endpoint
    except Exception as e:
        log.debug(f'Error reading http headers: {e}')


def _get_endpoint_in_html(response: requests.Response) -> Optional[str]:
    """Search for a webmention endpoint in HTML."""
    a_soup = BeautifulSoup(response.text, 'html.parser')

    # Check HTML <head> for <link> webmention endpoint
    try:
        links = a_soup.head.find_all('link', href=True, rel=True)
        for link in links:
            if 'webmention' in link['rel']:
                endpoint = link['href']
                log.debug(
                    f'webmention endpoint found in document head - '
                    f'address={endpoint}')
                return endpoint
    except Exception as e:
        log.debug(f'Error reading <head> of external link: {e}')

    # Check HTML <body> for <a> webmention endpoint
    try:
        links = a_soup.body.find_all('a', href=True, rel=True)
        for link in links:
            if 'webmention' in link['rel']:
                log.debug('webmention endpoint found in document body')
                endpoint = link['href']
                return endpoint
    except Exception as e:
        log.debug(f'Error reading <body> of link: {e}')


def _send_webmention(source_url: str, endpoint: str, target: str) -> Tuple[bool, int]:
    payload = {
        'target': target,
        'source': f'https://{settings.DOMAIN_NAME}{source_url}'
    }
    log.info(f'{endpoint}: {payload}')
    response = requests.post(endpoint, data=payload)
    status_code = response.status_code
    if status_code >= 300:
        log.warning(f'Sending webmention to "{endpoint}" '
                    f'FAILED with status_code={status_code}')
        return False, status_code
    else:
        log.info(f'Sending webmention to "{endpoint}" '
                 f'successful with status_code={status_code}')
        return True, status_code


def _relative_to_absolute_url(response: requests.Response, url: str) -> Optional[str]:
    """
    If given url is relative, try to construct an absolute url using response domain.
    """
    if not url:
        return None

    try:
        URLValidator()(url)
        return url  # url is already well-formed.
    except ValidationError:
        scheme, domain, _, _, _ = urlsplit(response.url)
        if not scheme or not domain:
            return None
        return f'{scheme}://{domain}/' \
            f'{url if not url.startswith("/") else url[1:]}'



"""
Tests for handling webmentions are sent to us from elsewhere.
"""

import logging

from django.http import HttpResponse
from django.test import TestCase, Client
from django.urls import reverse

from mentions.tasks import incoming_webmentions
from mentions.tests.util import (
    functions,
    constants,
)
from mentions.tests.models import MentionableTestModel
from mentions.util import split_url

log = logging.getLogger(__name__)

HTML_MENTION = '''
<html>
<head>
</head>
<body><a href="{target_url}">This mentions the target!</a></body>
</html>
'''


class MockHttpClient:
    """Wrap Django Client with API for `python-requests`."""

    def __init__(self, client: Client):
        self.django_client = client

    def get(self, url):
        return MockResponse(self.django_client.get(url))


class MockResponse:
    """Wrap Django HttpResponse with API for `python-requests`."""

    def __init__(self, response: HttpResponse):
        self.response = response
        self.headers = {
            'content-type': response._content_type_for_repr,
        }
        self.status_code = response.status_code
        self.text = response.content


class IncomingWebmentionsTests(TestCase):
    """"""

    def setUp(self):
        self.target_id, self.target_slug = functions.get_id_and_slug()
        self.target_url = functions.build_object_url(self.target_slug)

        target = MentionableTestModel.objects.create(
            stub_id=self.target_id,
            slug=self.target_slug,
            content='some html content',
            allow_incoming_webmentions=True)
        target.save()

    def test_get_target_path(self):
        """Ensure that path is retrieved from url correctly."""
        scheme, domain, path = split_url(self.target_url)
        self.assertEqual(reverse(constants.view_all_endpoints, args=[self.target_slug]), path)

    def test_get_target_object(self):
        """Ensure that database object is retrieved from url correctly."""
        log.debug(self.target_url)

        retrieved_model = incoming_webmentions._get_target_object(self.target_url)

        self.assertEqual(retrieved_model.slug, self.target_slug)

    def test_get_incoming_source(self):
        """Ensure that webmention source page can be retrieved correctly."""

        source_stub_id, source_slug = functions.get_id_and_slug()
        source = MentionableTestModel.objects.create(
            stub_id=source_stub_id,
            slug=source_slug,
            content=HTML_MENTION.format(target_url=self.target_url))
        source.save()
        source_url = functions.build_object_url(source_slug)

        self.assertIsNotNone(
            incoming_webmentions._get_incoming_source(
                source_url, client=MockHttpClient(self.client)))
